//============================================================================
//------------------------( T  O  U  R  B  O  T )-----------------------------
//============================================================================
// Overview: This program controls TourBot's webcam with pan/tilt servo tracking
// using data being feed by the Roborealm image processing software. It also
// controls the wheels for mobility and sonar for obstacle avoidance.
//============================================================================
//                             Begin Program
//============================================================================
// Variables
//============================================================================
// LIBRARIES
#include <Servo.h>                    // Include the servo library

// FUNCTION TIMING, Intervals in Milliseconds
unsigned long Interval1 = 30; unsigned long Previous1 = 0;  // ParseData,PanTilt,Wheels
unsigned long Interval2 = 45; unsigned long Previous2 = 0;  // Pan/tilt reset
unsigned long Interval3 = 3;  unsigned long Previous3 = 0;  // Wheel stop
unsigned long Interval4 = 1;  unsigned long Previous4 = 0;  // Pan/tilt
unsigned long Interval5 = 3;  unsigned long Previous5 = 0;  // Wheel

// PIN ALLOCATIONS
const byte PanPin = 6;                // Pin for Pan servo
const byte TiltPin = 7;               // Pin for Tilt servo
const byte LWheel = 11;               // Pin for wheel forward/backwards mixed mode PWM to motorcontroller S2
const byte RWheel = 10;               // Pin for wheel turning mixed mode PWM to motorcontroller S1

// OPTICS
const byte Repeats = 15;              // Max number of ([COG_X]=0) to keep target locked in view
const byte zMinAbs = 7;               // [COG_BOX_SIZE] absolute minimum to stop following
const byte zMin = 8;                  // [COG_BOX_SIZE] min
const byte zFixed = 15;               // [COG_BOX_SIZE] set point to stand still
const byte zMax = 24;                 // [COG_BOX_SIZE] max
const int Width = 320;                // Image width
const int Height = 240;               // Image height
const int CenterX = Width/2;          // Image vertical center
const int CenterY = Height/2;         // Image horizontal center
int var[] = {0,0,0};                  // Array to store parsed data sequence from Roborealm

// PAN/TILT
Servo PanServo;                       // Servo for pan webcam
Servo TiltServo;                      // Servo for tilt webcam
const byte PanMax = 150;              // Pan angle max range
const byte PanMin = 30;               // Pan angle min range
const byte TiltMax = 130;             // Tilt angle max range
const byte TiltMin = 50;              // Tilt angle min range
const byte PTcenter = 90;             // Center angle for Pan/Tilt servos
const byte KpX = 6;                   // Pan porportional gain
const byte KiX = 0;                   // Pan integral gain
const byte KdX = 1;                   // Pan derivative gain
const byte KpY = 5;                   // Tilt porportional gain
const byte KiY = 0;                   // Tilt integral gain
const byte KdY = 1;                   // Tilt derivative gain
int PanAngleSet = PTcenter;           // Angle new for Pan servo
int TiltAngleSet = PTcenter;          // Angle new for Tilt servo
int PanAngle = PanAngleSet;           // PanServo angle initialized to center (90 degrees)
int TiltAngle = TiltAngleSet;         // TiltServo angle initialized to center (90 degrees)
int ErrX, ErrLastX, DerivativeX, IntegralX; // Pan PID variables
int ErrY, ErrLastY, DerivativeY, IntegralY; // Tilt PID variables

// WHEELS
const byte WheelPanAngleMax = 140;    // Maximum pan angle for wheel turning to respond
const byte WheelPanAngleMin = 40;     // Minimum pan angle for wheel turning to respond
const byte LWheelMax = 255;           // Maxium pwm for left wheel forward
const byte LWheelMin = 0;             // Minium pwm for left wheel reverse
const byte RWheelMax = 241;           // Maxium pwm for left wheel forward
const byte RWheelMin = 15;            // Minium pwm for right wheel reverse
const byte WheelStop = 128;           // PWM to stop left/right wheels
int LWheelPWMSet = WheelStop;         // PWM new for left wheel
int RWheelPWMSet = WheelStop;         // PWM new for right wheel
int LWheelPWM = LWheelPWMSet;         // PWM current value for left wheel
int RWheelPWM = RWheelPWMSet;         // PWM current value for right wheel

// SONAR
const byte SonarRangeMax = 25;        // Maximum range for sonar to detect
const byte SonarRangeMin = 8;         // Minimum range for sonar to detec
long Inches[] = {0,0,0,0,0,0};        // Array containing sonar inches for each sonar

// COUNTERS AND FLAGS
int Count = 0;                        // Count for data repeats counter
//============================================================================
// Setup initial conditions
//============================================================================
void setup(){
// SERIAL PORTS
  Serial.begin(115200);               // Open serial port
  Serial1.begin(115200);              // Open serial port

// PAN/TILT
  PanServo.attach(PanPin);            // Attach panServo to pin
  TiltServo.attach(TiltPin);          // Attach tiltServo to pin
  PanServo.write(PanAngle);           // Initialize panServo to 90
  TiltServo.write(TiltAngle);         // Initialize tiltServo to 90

// WHEELS
  pinMode(LWheel, OUTPUT);            // Sets the pin as output
  pinMode(RWheel, OUTPUT);            // Sets the pin as output
  analogWrite(LWheel, LWheelPWM);     // Initialize LeftWheel to PWM 128
  analogWrite(RWheel, RWheelPWM);     // Initialize RightWheel to PWM 128
}
//============================================================================
// Main Loop
//============================================================================
void loop(){
  if(millis() - Previous1 >= Interval1){
    Previous1 = millis();
    ParseOpticsData();                          // Parse data from Roborealm
    ParseSonarData();                           // Parse data from Roborealm
    PrintData();                                // Print out data for analysis
    if(Count != Repeats && var[0] > 0){
      PanTilt();                                // Get angles to set for pan/tilt servos
      Wheels();                                 // Get pwms to set for wheels 
    }
  }
  if(Count == Repeats){
    if(var[0] == 0){
      if(millis() - Previous2 >= Interval2){
        Previous2 = millis();
        ResetPanTilt();                         // Reset pan/tilt servos
        UpdatePanTilt();                        // Update current pan/tilt angle
      }
    }
    if(millis() - Previous3 >= Interval3){
      Previous3 = millis();
      StopWheels();                             // Bring wheels to a stop
      UpdateWheels();                           // Update current wheel pwms
    }
  }
  else{
    if(millis() - Previous4 >= Interval4){
      Previous4 = millis();
      UpdatePanTilt();                          // Update current pan/tilt angle
    }
    if(millis() - Previous5 >= Interval5){
      Previous5 = millis();
      UpdateWheels();                           // Update current wheel pwms
    }
  }
}
//============================================================================
////////////////////////////////// Functions ///////////////////////////////// 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Parse the Roborealm serial sequence: " <[COG_X];[COG_Y];[COG_BOX_SIZE];> "
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void ParseOpticsData(){
  if(Serial.available() > 0){          // If data available in serial buffer, begin reading in characters
    Count = 0;                         // Reset repeats count
    byte i = 0;
    byte Flag = 0;
    int pos = 0;
    char header = Serial.read();
    while(header == '<'){              // If the header byte '<' seen, begin parsing data in serial buffer
      char ch = Serial.read();
      if(ch >= '0' && ch <= '9' && Flag == 0){      // If character(ch) read is a number, accumulate the value
        pos = pos * 10 + ch - '0';     // Convert character number sequence to integer equivalent
      }
      else if(ch == '.'){
        Flag = 1;
      }
      else if(ch == ';'){              // Store value into array when delimiter byte ';' seen
        var[i] = pos;
        pos = Flag = 0;
        i++;
      }
      else if(ch == '>'){              // Break the parsing loop when the ending byte '>' is seen
        header = ch;
      }
    }
  }
  else if(Count != Repeats){
    Count++;                           // Count repetitive data
  }
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Parse data sequence from microcontroller controlling sonar: 
// " <FrontLeft,FrontCenter,FrontRight,RearLeft,RearCenter,RearRight;> "
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void ParseSonarData(){
  if(Serial1.available() > 0){         // If data available in serial buffer, begin reading in characters
    byte i = 0;
    int pos = 0;
    char header = Serial1.read();
    while(header == '<'){              // If the header byte '<' seen, begin parsing data in serial buffer
      char ch = Serial1.read();
      if(ch >= '0' && ch <= '9'){      // If character(ch) read is a number, accumulate the value
        pos = pos * 10 + ch - '0';     // Convert character number sequence to integer equivalent
      }
      else if(ch == ';'){              // Store value into array when delimiter byte ';' seen
        Inches[i] = pos;
        pos = 0;
        i++;
      }
      else if(ch == '>'){              // Break the parsing loop when the ending byte '>' is seen
        header = ch;
      }
    }
  }
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Pan/Tilt servo driver using PID
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void PanTilt(){
// Pan PID
  ErrX = constrain(CenterX - var[0],-CenterX,CenterX);        // Pan Proportional term
  IntegralX = constrain(IntegralX + ErrX,-CenterX,CenterX);   // Pan Integral term
  DerivativeX = constrain(ErrX - ErrLastX,-CenterX,CenterX);  // Pan Derivative term
  PanAngleSet = constrain(PanAngleSet - (ErrX*KpX + IntegralX*KiX + DerivativeX*KdX)/CenterX,PanMin,PanMax);;
  ErrLastX = ErrX;
// Tilt PID
  ErrY = constrain(CenterY - var[1],-CenterY,CenterY);        // Tilt Proportional term
  IntegralY = constrain(IntegralY + ErrY,-CenterY,CenterY);   // Tilt Integral term
  DerivativeY = constrain(ErrY - ErrLastY,-CenterY,CenterY);  // Tilt Derivative term
  TiltAngleSet = constrain(TiltAngleSet - (ErrY*KpY + IntegralY*KiY + DerivativeY*KdY)/CenterY,TiltMin,TiltMax);
  ErrLastY = ErrY;
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Pan/Tilt reset
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void ResetPanTilt(){
  IntegralX = ErrLastX = IntegralY = ErrLastY = 0;  // Reset PID varialbes
  PanAngleSet = Current(PanAngleSet,PTcenter);      // Bring pan servo to 90 degrees
  TiltAngleSet = Current(TiltAngleSet,PTcenter);    // Bring tilt servo to 90 degrees
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Wheel driver, controlled by Roborealm and sonar
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void Wheels(){
  if(var[2] >= zMinAbs){
    LWheelPWMSet = map(PanAngleSet,WheelPanAngleMin,WheelPanAngleMax,LWheelMin,LWheelMax);      // Turn left
    RWheelPWMSet = map(PanAngleSet,WheelPanAngleMin,WheelPanAngleMax,RWheelMax,RWheelMin);      // Turn right
    LWheelPWMSet = constrain(LWheelPWMSet,LWheelMin,LWheelMax);
    RWheelPWMSet = constrain(RWheelPWMSet,RWheelMin,RWheelMax);
    if(var[2] < zFixed){
      LWheelPWMSet = map(var[2],zFixed,zMin,WheelStop,LWheelMax) - (WheelStop - LWheelPWMSet);  // Left forward
      RWheelPWMSet = map(var[2],zFixed,zMin,WheelStop,RWheelMax) - (WheelStop - RWheelPWMSet);  // Right forward
      LWheelPWMSet = constrain(LWheelPWMSet, LWheelMin, LWheelMax);
      RWheelPWMSet = constrain(RWheelPWMSet, RWheelMin, RWheelMax);
      if(Inches[0] > 0 && Inches[0] < SonarRangeMax){
        int FL = map(Inches[0],SonarRangeMin,SonarRangeMax,WheelStop,RWheelPWMSet);             // Check front left sonar
        RWheelPWMSet = max(FL,WheelStop);
      }
      if(Inches[2] > 0 && Inches[2] < SonarRangeMax){
        int FR = map(Inches[2],SonarRangeMin,SonarRangeMax,WheelStop,LWheelPWMSet);             // Check front right sonar
        LWheelPWMSet = max(FR,WheelStop);
      }
      if(Inches[1] > 0 && Inches[1] < SonarRangeMax){
        LWheelPWMSet = map(Inches[1],SonarRangeMin,SonarRangeMax,WheelStop,LWheelPWMSet);       // Check front center sonar affect on front left sonar
        RWheelPWMSet = map(Inches[1],SonarRangeMin,SonarRangeMax,WheelStop,RWheelPWMSet);       // Check front center sonar affect on front right sonar
        LWheelPWMSet = max(LWheelPWMSet,WheelStop);
        RWheelPWMSet = max(RWheelPWMSet,WheelStop);
      }
    }
    else if(var[2] > zFixed){
      LWheelPWMSet = map(var[2],zFixed,zMax,WheelStop,LWheelMin) - (WheelStop - LWheelPWMSet);  // Left reverse
      RWheelPWMSet = map(var[2],zFixed,zMax,WheelStop,RWheelMin) - (WheelStop - RWheelPWMSet);  // Right reverse
      LWheelPWMSet = constrain(LWheelPWMSet, LWheelMin, LWheelMax);
      RWheelPWMSet = constrain(RWheelPWMSet, RWheelMin, RWheelMax);
      if(Inches[3] > 0 && Inches[3] < SonarRangeMax){
        int RL = map(Inches[3],SonarRangeMin,SonarRangeMax,WheelStop,RWheelPWMSet);             // Check rear left sonar       
        RWheelPWMSet = min(RL,WheelStop);
      }
      if(Inches[5] > 0 && Inches[5] < SonarRangeMax){
        int RR = map(Inches[5],SonarRangeMin,SonarRangeMax,WheelStop,LWheelPWMSet);             // Check rear right sonar
        LWheelPWMSet = min(RR,WheelStop);
      }
      if(Inches[4] > 0 && Inches[4] < SonarRangeMax){
        LWheelPWMSet = map(Inches[4],SonarRangeMin,SonarRangeMax,WheelStop,LWheelPWMSet);       // Check rear center sonar affect on rear left sonar
        RWheelPWMSet = map(Inches[4],SonarRangeMin,SonarRangeMax,WheelStop,RWheelPWMSet);       // Check rear center sonar affect on rear right sonar
        LWheelPWMSet = min(LWheelPWMSet,WheelStop);
        RWheelPWMSet = min(RWheelPWMSet,WheelStop);
      }
    }
  }
  else{
    LWheelPWMSet = RWheelPWMSet = WheelStop;
  }
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Slowly bring wheels to stop
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void StopWheels(){
  LWheelPWMSet = Current(LWheelPWMSet,WheelStop);        // Bring left wheel to stop
  RWheelPWMSet = Current(RWheelPWMSet,WheelStop);        // Bring right wheel to stop
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Takes current value and brings closer to a set value at a defined rate
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int Current(int Num, int NumSet){
  if(Num != NumSet){
    Num += (NumSet - Num)/max(abs(NumSet - Num),1);      // Increase/decrease by 1 at set rate in (ms)
  }
  else{
    Num = NumSet;
  }
  return Num;
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Update Pan/Tilt
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void UpdatePanTilt(){
// Pan
  PanAngle = Current(PanAngle,PanAngleSet);
  PanServo.write(PanAngle);
// Tilt
  TiltAngle = Current(TiltAngle,TiltAngleSet);
  TiltServo.write(TiltAngle);
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Update Wheels
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void UpdateWheels(){
// Left wheel
  LWheelPWM = Current(LWheelPWM,LWheelPWMSet);
  analogWrite(LWheel,LWheelPWM);
// Right wheel
  RWheelPWM = Current(RWheelPWM,RWheelPWMSet);
  analogWrite(RWheel,RWheelPWM);
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Print outputs for analysis 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void PrintData(){
  Serial1.print('<');
  Serial1.print(var[2]);Serial1.print(';');
  Serial1.print(Count);Serial1.print(';');
  Serial1.print(int(Repeats));Serial1.print(';');
  Serial1.print(int(zMinAbs));Serial1.print(';');
  Serial1.print(int(zFixed));Serial1.print(';');
  Serial1.print(int(SonarRangeMax));Serial1.print(';');
  Serial1.print('>');
//  char command[] = {'X','Y','Z','FLSonar','FCSonar','FRSonar'};
//  for (int i = 0; i < 3; i++){
//    Serial.print(command[i]);Serial.print(var[i]);Serial.print(' ');
//  }
//  Serial.println();
//  Serial.print("PanAngle:");Serial.print(PanAngle);
//  Serial.print(" TiltAngle:");Serial.print(TiltAngle);
//  Serial.println();
//  Serial.print("LWheel:");Serial.print(LWheelPWM);
//  Serial.print(' ');
//  Serial.print("RWheel:");Serial.print(RWheelPWM);
//  Serial.println();
//  Serial.println();
}
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//////////////////////////////////////////////////////////////////////////////
//============================================================================
//                             End Program
//============================================================================

